# This file verifies that all principal congruence link complements with class
# number 1 for which we do not have a link complement are indeed link
# complements. More precisely, it checks that each triangulation listed below
# directory becomes the 3-sphere when Dehn-filled along all meridians (1,0).
#
# Usage:
#     python proveLinkComplements.py
# or:
#     sage -python proveLinkComplements.py
#
# It requires SnapPy to be installed in your python, respectively, Sage
# environment.
#
#
# We call the triangulation files in this directory
# LINK COMPLEMENT CERTIFICATES.
# They allow one to prove that a principal congruence manifold is a link
# complement by 
# 1. Recreating the principal congruence manifold
#    (using generatePrincipalCongruenceManifold.py and reginaToSnapPy.py).
# 2. Check that the result is isometric to the file in this directory
#    (using SnapPy's Manifold.is_isometric_to).
# 3. Check that the file in the directory becomes the 3-sphere when Dehn-filling
#    along all meridians (which is what this script is doing).
#
# Matthias Goerner, enischte@gmail.com, 10/20/2017

from snappy import Triangulation

def prove_link_complement(filename):

    # Open file
    M = Triangulation(filename)

    # Remember number of cusps
    num_cusps = M.num_cusps()
    
    # Fill cusps along meridian
    while M.num_cusps() > 1:
        # Process 20 at a time, but skip the last one
        n = min(20, M.num_cusps() - 1)
        M.dehn_fill(n * [(1,0)])
        # Replace triangulation by filled triangulation.
        # We do this since filling and simplifying the triangulation appears to
        # be cheaper than simplifying the presentation of the fundamental group.
        M = M.filled_triangulation()

    # Dehn-fill the last cusp, but do leave the triangulation with one
    # ideal vertex. This, in general, has less tetrahedra than a triangulation
    # that has only finite vertices.
    M.dehn_fill((1,0))

    # Check that SnapPy could show it is the trivial group.
    G = M.fundamental_group() 
    if G.generators():
        raise Exception(
            "%s: SnapPy failed to show that the fundamental group is trivial: %s" % (
                filename, G))
    else:
        print "%s is a %d-link complement" % (filename, num_cusps)

def main():
    # All principal congruence link complements with class number 1 for which we
    # do not have a link diagram.
    discriminant1 = [
        "pSL_2_plus_0_sqrt_minus_1.tri",
        "pSL_2_plus_1_sqrt_minus_1.tri",
        "pSL_2_plus_2_sqrt_minus_1.tri",
        "pSL_3_plus_0_sqrt_minus_1.tri",
        "pSL_3_plus_1_sqrt_minus_1.tri",
        "pSL_3_plus_2_sqrt_minus_1.tri",
        "pSL_4_plus_1_sqrt_minus_1.tri"
        ]
        
    discriminant2 = [
        "pSL_1_plus_1_sqrt_minus_2.tri",
        "pSL_1_plus_2_sqrt_minus_2.tri",
        "pSL_2_plus_0_sqrt_minus_2.tri",
        "pSL_2_plus_1_sqrt_minus_2.tri",
        "pSL_3_plus_1_sqrt_minus_2.tri"
        ]

    discriminant3 = [
        "pSL_2_plus_0_sqrt_minus_3.tri",
        "pSL_2.5_plus_0.5_sqrt_minus_3.tri",
        "pSL_3_plus_0_sqrt_minus_3.tri",
        "pSL_3_plus_1_sqrt_minus_3.tri",
        "pSL_3.5_plus_0.5_sqrt_minus_3.tri",
        "pSL_4_plus_1_sqrt_minus_3.tri",
        "pSL_4.5_plus_0.5_sqrt_minus_3.tri"
        ]

    discriminant5 = [
        "pSL_3_plus_0_sqrt_minus_5_and_1_plus_1_sqrt_minus_5.tri"
        ]

    discriminant7 = [
        "pSL_0_plus_1_sqrt_minus_7.tri",
        "pSL_0.5_plus_0.5_sqrt_minus_7.tri",
        "pSL_0.5_plus_1.5_sqrt_minus_7.tri",
        "pSL_1_plus_1_sqrt_minus_7.tri",
        "pSL_1.5_plus_0.5_sqrt_minus_7.tri",
        "pSL_2_plus_0_sqrt_minus_7.tri",
        "pSL_2_plus_1_sqrt_minus_7.tri",
        "pSL_2.5_plus_0.5_sqrt_minus_7.tri",
        "pSL_3.5_plus_0.5_sqrt_minus_7.tri"
        ]

    discriminant11 = [
        "pSL_0.5_plus_0.5_sqrt_minus_11.tri",
        "pSL_1.5_plus_0.5_sqrt_minus_11.tri",
        "pSL_2.5_plus_0.5_sqrt_minus_11.tri"
        ]

    discriminant15 = [
        "pSL_2_plus_0_sqrt_minus_15_and_0.5_plus_0.5_sqrt_minus_15.tri",
        "pSL_3_plus_0_sqrt_minus_15_and_1.5_plus_0.5_sqrt_minus_15.tri",
        "pSL_4_plus_0_sqrt_minus_15_and_0.5_plus_0.5_sqrt_minus_15.tri",
        "pSL_5_plus_0_sqrt_minus_15_and_2.5_plus_0.5_sqrt_minus_15.tri",
        "pSL_6_plus_0_sqrt_minus_15_and_-1.5_plus_0.5_sqrt_minus_15.tri"
        ]

    discriminant19 = [
        "pSL_0.5_plus_0.5_sqrt_minus_19.tri"
        ]

    discriminant23 = [
        "pSL_2_plus_0_sqrt_minus_23_and_0.5_plus_0.5_sqrt_minus_23.tri",
        "pSL_3_plus_0_sqrt_minus_23_and_0.5_plus_0.5_sqrt_minus_23.tri",
        "pSL_4_plus_0_sqrt_minus_23_and_-1.5_plus_0.5_sqrt_minus_23.tri"
        ]

    discriminant31 = [
        "pSL_2_plus_0_sqrt_minus_31_and_0.5_plus_0.5_sqrt_minus_31.tri",
        "pSL_4_plus_0_sqrt_minus_31_and_0.5_plus_0.5_sqrt_minus_31.tri",
        "pSL_5_plus_0_sqrt_minus_31_and_1.5_plus_0.5_sqrt_minus_31.tri"
        ]

    discriminant47 = [
        "pSL_2_plus_0_sqrt_minus_47_and_0.5_plus_0.5_sqrt_minus_47.tri",
        "pSL_3_plus_0_sqrt_minus_47_and_0.5_plus_0.5_sqrt_minus_47.tri",
        "pSL_4_plus_0_sqrt_minus_47_and_0.5_plus_0.5_sqrt_minus_47.tri"
        ]

    discriminant71 = [
        "pSL_2_plus_0_sqrt_minus_71_and_0.5_plus_0.5_sqrt_minus_71.tri"
        ]

    for filename in (discriminant1  +
                     discriminant2  +
                     discriminant3  +
                     discriminant5  +
                     discriminant7  +
                     discriminant11 +
                     discriminant15 +
                     discriminant19 +
                     discriminant23 +
                     discriminant31 +
                     discriminant47 +
                     discriminant71):
        prove_link_complement(filename)

if __name__ == '__main__':
    main()
