# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

"""
A library to create Bianchi orbifolds by computing the Dirichlet domain for
a Bianchi group.

A Bianchi orbifold can be created with generateBinachiOrbifold. If it is saved
into the sub-directory data under the name given by filenameForBianchiOrbifold,
it can be easily loaded as loadBianchiOrbifold.

All information about the Bianchi orbifold is packaged into a regina container
and can be accessed through OrbifoldData.
"""

try:
    import sage.all
except:
    raise Exception(
        "Requires SageMath (http://www.sagemath.org/). "
        "Run as 'sage -python generateBianchiOrbifold.py'.")

try:
    import regina
except:
    raise Exception(
        "Requires sageRegina (http://sageregina.unhyperbolic.org/)")

from .imagQuadraticFieldUtil import *
from .ideal import *
from .pslMatrix import *
from .dirichletDomain import *
from .bianchiGroupSampler import *
from .dirichletDomainTriangulator import *
from .orbifoldData import *

from .data import __path__ as _data_paths

import os

def generateBianchiOrbifold(D, range_a, range_d):
    """
    Generate the orbifold associated to the Bianchi group PSL(2, O) where
    O is the ring of integers of Q(sqrt(-D)). range_a and range_d are integers
    controlling how many matrices from the Bianchi group are used to compute
    the Dirichlet domain, see BianchiGroupSampler. 2 and 2 is is large enough to
    generate the Bianchi groups for D = 1, 2, 3, 7, 11, 19.

    The result is a regina container bundling all the information about the
    Bianchi orbifold, see OrbifoldData for more detail.
    """

    quadField = ImagQuadraticFieldUtil(D)
    g = BianchiGroupSampler(quadField, range_a, range_d)
    d = DirichletDomain(g)
    return DirichletDomainTriangulator(d).container

def filenameForBianchiOrbifold(D):
    """
    The preferred filename for storing the regina container with the
    Bianchi orbifold.
    """

    return 'Bianchi%r.rga' % D

def loadBianchiOrbifold(D):
    """
    Load the Bianchi orbifold for Q(sqrt(-D)) from the data directory.
    The result is a OrbifoldData object.
    """

    filename = filenameForBianchiOrbifold(D)
    path = os.path.join(_data_paths[0], filename)

    c = regina.open(path)
    if not c:
        raise Exception(
            ("Could not find orbifold file %s in directory %s. "
             "You can create a Bianchi orbifold for the desired discrimnant %d "
             "with generateBianchiOrbifold.py and move it to the directory." % (
                    filename, _data_paths[0], D)))
    
    return OrbifoldData(c)

