# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from .polyhedronUtils import BarycentricSimplex
from .pslMatrix import *

import regina

__all__ = ['DirichletDomainTriangulator']

class DirichletDomainTriangulator:
    """
    Given a DirichletDomain d, produces a Regina triangulation of the
    orbifold obtained by gluing paired faces and packages it into a Regina
    container together with various other pieces of information about the
    Dirichlet domain, also see OrbifoldData for further details.
    
    This container can be obtained by
    >>> c = DirichletDomainTriangulator(d).container

    Such a container can be unpackaged by OrbifoldData which provides access
    to the deserialized information about the orbifold.

    The container can be saved to disk with c.save("FILE.rga") and retrieved
    and unpacked later with OrbifoldData(regina.open("FILE.rga")).
    """

    def _compute_triangulation_and_matrices(self):
        # Corresponds to OrbifoldData.triangulation
        self.triangulation = regina.Triangulation3()
        self.triangulation.setLabel('Triangulation')
        
        # Corresponds to OrbifoldData.matrices
        self.matrices = []

        # We create a new regina simplex for each BarycentricSimplex
        # and store it here.
        simplexToReginaSimplex = {}

        face_and_edge_to_simplices = {}
        face_and_vertex_to_simplices = {}
        edge_and_vertex_to_simplices = {}

        for s in self.dirichletDomain.barycentricSimplices:
            # face, edge, and vertex of polyhedron this simplex is adjacent to
            face, edge, vertex = s
            face_and_edge_to_simplices.setdefault((face, edge),[]).append(s)
            face_and_vertex_to_simplices.setdefault((face, vertex),[]).append(s)
            edge_and_vertex_to_simplices.setdefault((edge, vertex),[]).append(s)
        
            simplexToReginaSimplex[s] = self.triangulation.newSimplex()
            self.matrices.append(self.dirichletDomain.face_to_matrix[face])

        # Two barycentric simplices adjacent to the same face and edge of
        # the polyhedron are glued along face 0
        for simp0, simp1 in face_and_edge_to_simplices.values():
            simplexToReginaSimplex[simp0].join(
                0, simplexToReginaSimplex[simp1], regina.NPerm4())
            
        # Similar for face and vertex and face 1 
        for simp0, simp1 in face_and_vertex_to_simplices.values():
            simplexToReginaSimplex[simp0].join(
                1, simplexToReginaSimplex[simp1], regina.NPerm4())

        # Similar for edge and vertex and face 2
        for simp0, simp1 in edge_and_vertex_to_simplices.values():
            simplexToReginaSimplex[simp0].join(
                2, simplexToReginaSimplex[simp1], regina.NPerm4())
        
        # It is left to glue face 3 of each barycentric simplex
            
        # Iterate through barycentric simplex and corresponding regina simplex
        # simultaneously
        for s, reginaSimp in simplexToReginaSimplex.items():
            # face, edge, and vertex of polyhedron this simplex is adjacent to
            face, edge, vertex = s

            # find the mate face and the correspondence between vertices of
            # the face and the mate face
            opposite_face, vertex_to_opposite_vertex = (
                self.dirichletDomain.face_to_opposite_face_and_vertex_to_opposite_vertex[face])

            # find the corresponding edge on the mate face
            opposite_edge = tuple(
                sorted([ vertex_to_opposite_vertex[v] for v in edge]))

            # find the corresponding vertex of this vertex
            opposite_vertex = vertex_to_opposite_vertex[vertex]

            # we can now see what barycentric simplex corresponds to this
            # barycentric simplex under the face-pairing
            opposite_simp = BarycentricSimplex((
                opposite_face, opposite_edge, opposite_vertex))

            # And glue them along face 3
            reginaSimp.join(3, simplexToReginaSimplex[opposite_simp], regina.NPerm4())

    def _compute_orders(self):
        identity = PslMatrix.identity(self.dirichletDomain.imagQuadraticFieldUtil)

        def compute_matrix_for_edge(startSimp, edge_index):
            """
            Computes the matrix when going around the specified edge of the
            specified simplex.
            """
            m = identity
            simp = startSimp

            while True:
                # Keep multiplying face-pairing matrices
                m = m * self.matrices[simp.index()]
                # Advance about the edge
                simp = simp.adjacentSimplex(3).adjacentSimplex(edge_index)
                
                if simp == startSimp:
                    return m

        def compute_order(reginaSimp, edge_index):
            """
            Computes the order of the matrix obtained by going around the specified
            edge of the specified simplex. This is the order of the edge as singular
            locus.
            """

            edge_matrix = compute_matrix_for_edge(reginaSimp, edge_index)
            
            order = 0

            m = identity
            while True:
                m = m * edge_matrix
                order += 1
                    
                if m == identity:
                    return order

        # Corresponds to OrbifoldData.orders
        self.orders = [
            [ compute_order(s, edge_index) for edge_index in range(3) ]
              for s in self.triangulation.simplices() ]

    def __init__(self, dirichletDomain):
        self.dirichletDomain = dirichletDomain
        
        self._compute_triangulation_and_matrices()
        self._compute_orders()

        # Corresponds to OrbifoldData.disc
        self.disc = regina.Text('%d' % self.dirichletDomain.imagQuadraticFieldUtil.D)
        self.disc.setLabel('D')

        # self.matrices serialized to Text packet
        self.reginaMatrices = regina.Text('%r' % self.matrices)
        self.reginaMatrices.setLabel('Matrices')

        # self.orders serialized to Text packet
        self.reginaOrders = regina.Text('%r' % self.orders)
        self.reginaOrders.setLabel('Orders')

        # Corresponds to OrbifoldData.container
        self.container = regina.Container('TriangulationWithMatrices')
        self.container.insertChildLast(self.triangulation)
        self.container.insertChildLast(self.disc)
        self.container.insertChildLast(self.reginaMatrices)
        self.container.insertChildLast(self.reginaOrders)
