# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from sage.all import RationalField, QuadraticField, AA, QQ, QuaternionAlgebra, sqrt, factor, prod, ceil

__all__ = ['ImagQuadraticFieldUtil']

class ImagQuadraticFieldUtil:
    """
    Provides various values and functions around Sage's QuadraticField(-D)
    where D is a square-free positive number.
    See __init__ for details.
    """

    def __init__(self, D):
        # D is assumed to be Square-free
        self.D = D

        try:
            squarefree = self.D.is_squarefree()
        except:
            raise Exception(
                "ImageQuadraticFieldUtil(%r): D must be a Sage Integer" % D)

        if not squarefree:
            raise Exception(
                "ImageQuadraticFieldUtil(%r): D must be square-free" % D)

        if not D > 0:
            raise Exception(
                "ImageQuadraticFieldUtil(%r): D must be positive" % D)
                
        #
        # Setup the Complex Number Field Q(sqrt(-D))

        # The name of the generator of Q(sqrt(-D))
        self.sqrtMinusDName = 'sqrtMinus%d' % D
        # The number field Q(sqrt(-D))
        self.CNF = QuadraticField(-D, self.sqrtMinusDName)
        # The generator sqrt(-D) in Q(sqrt(-D))
        self.sqrtMinusD = self.CNF.gen()
        
        # Generates the ring of integers of Q(sqrt(-D)) as a
        # lattice together with 1, i.e.,
        # the ring of integers is Z + Z * algInt.
        if -D % 4 == 1:
            self.algInt = (1 + self.sqrtMinusD) / 2
        else:
            self.algInt = self.sqrtMinusD
            
        # All units in the ring of integers
        if D == 1:
            self.units = [
                1, -1,
                self.sqrtMinusD, -self.sqrtMinusD ]
        elif D == 3:
            self.units = [
                1, -1, 
                ( 1+self.sqrtMinusD)/2, (-1-self.sqrtMinusD)/2,
                (-1+self.sqrtMinusD)/2, ( 1-self.sqrtMinusD)/2]
        else:
            self.units = [1, -1]
        
        # The corresponding real quadratic field Q(sqrt(D))
        # and sqrt(D) in Q(sqrt(D))
        if D == 1:
            # For D = 1, this is just the rationals.
            self.RNF = QQ
            self.sqrtD = 1
        else:
            self.RNF = QuadraticField(D)
            self.sqrtD = self.RNF.gen()
        
        # Quaternions a + b * i + c * j + d * k
        # where a, b, c, d are in Q(sqrt(D)).
        #
        # Q(sqrt(-D)) embeds into these quaternions via
        # to_quadratic_field_quaternion
        self.quadraticFieldQA = QuaternionAlgebra(self.RNF, -1, -1)
        
        # Quaternions a + b * i + c * j + d * k
        # where a, b, c, d can be any algebraic numbers in R.
        self.algebraicQA = QuaternionAlgebra(AA, -1, -1)
        # The sqrt(D) in the above quaternions.
        self.sqrtDinAA = AA(sqrt(D))

        # Is the ring of integers a PID/UFD?
        self.isPID = (self.D in [1, 2, 3, 7, 11, 19, 43, 67, 163])

        # Cache for factorization
        self._all_products_cache = {}

        # Cache for factorization in non-PIDs
        self._norm_cache_limit = 0
        self._norm_cache = {}

    def __repr__(self):
        return "ImagQuadraticFieldUtil(%d)" % self.D
       
    def _is_algebraic_integer_from_coefficients(self, a, b):
        """
        Returns true if a + b * sqrt(-D) is an algebraic integer.
        """

        if -self.D % 4 == 1:
            two_a = 2 * a
            two_b = 2 * b
            return (two_a.is_integer() and two_b.is_integer() and
                    two_a % 2 == two_b % 2)
        else:
            return a.is_integer() and b.is_integer()
 
    def _is_algebraic_integer(self, z):
        """
        Returns true if z in self.CNF is an algebraic integer.
        """

        return self._is_algebraic_integer_from_coefficients(
            *ImagQuadraticFieldUtil.getCoefficientsOfElement(z))

    def getAlgebraicInteger(self, a, b):
        """
        Given a and b, returns a + b * sqrt(-D) in Q(sqrt(-D)) and checks
        that it is an algebraic integer.
        a and b are assumed to be Sage Integer or Rational's.
        """

        if not self._is_algebraic_integer_from_coefficients(a, b):
            raise Exception(
                "Not an algebraic integer %r + %r * sqrt(%r)" % (
                    a, b, -self.D))

        return a + b * self.sqrtMinusD
        
    @staticmethod
    def _splittings_of_tuple(t):
        """
        Given (1,1,2), returns
        [((0,0,0), (1,1,2)),
         ((0,0,1), (1,1,1)),
         ((0,0,2), (1,1,0)),
         ((0,1,0), (1,0,2)),
         ...,
         ((1,1,1), (0,0,1)),
         ((1,1,2), (0,0,0))]
        i.e., all pairs of tuples that add to the given tuple element-wise.
        """
        
        if t:
            for e1 in range(t[0]+1):
                e2 = t[0] - e1
                for split1, split2 in ImagQuadraticFieldUtil._splittings_of_tuple(t[1:]):
                    yield ((e1,)+split1, (e2,)+split2)
        else:
            yield ((),())
            
    def _all_products_pid(self, z):
        """
        Handles the PID case of all_products.
        """

        # Factor z into primes
        fs = factor(z)
        terms = [f[0] for f in fs]
        degs  = [f[1] for f in fs]
        
        # Multiply powers of primes
        for degs1, degs2 in ImagQuadraticFieldUtil._splittings_of_tuple(degs):
            res = [ prod([term ** deg for term, deg in zip(terms, this_degs)]) 
                    for this_degs in [degs1, degs2]]
            for unit in self.units:
                yield (res[0] * fs.unit() * unit, res[1] / unit)
        
    def _all_algebraic_integers_with_norm(self, n):
        """
        Returns all algebraic integers with the given norm.
        """

        if n > self._norm_cache_limit:
            # Recreate the cache if we have not computed up to this norm yet
            self._norm_cache = {}
            for z in self.enumerate_alg_integers(ceil(sqrt(n)+1)):
                self._norm_cache.setdefault(z.norm(), []).append(z)
            self._norm_cache_limit = n

        return self._norm_cache.get(n, [])

    def _all_products_non_pid(self, z):
        """
        Handle non-PID case of all_products.
        This might return the same pair (u, v) with u * v = z multiple times.
        """

        n = z.norm()
        sqrt_n = ceil(sqrt(n))

        # Factor the norm |z|^2
        fs = factor(n)
        terms = [f[0] for f in fs]
        degs  = [f[1] for f in fs]
        for degs1, degs2 in ImagQuadraticFieldUtil._splittings_of_tuple(degs):
            res = prod([term ** deg for term, deg in zip(terms, degs1)])
            # For every factor of |z|^2 that is less than |z|
            # (since for one of u or v with u * v = z we must have |_| <= |z|)
            if res <= sqrt_n:
                # Get all algebraic integers of that norm
                for u in self._all_algebraic_integers_with_norm(res):
                    v = z / u
                    if self._is_algebraic_integer(v):
                        yield (u,v)
                        yield (v,u)

    def all_products(self, z):
        """
        Given an algebraic integer z in Q(sqrt(-D)), returns all pairs (u, v) of
        algebraic integers such that u * v = z.
        """

        if z == 0:
            raise Exception("Trying to factor 0")

        if not self._all_products_cache.has_key(z):
            # If not cached yet
            if self.isPID:
                self._all_products_cache[z] = sorted(
                    self._all_products_pid(z))
            else:
                self._all_products_cache[z] = sorted(
                    set(self._all_products_non_pid(z)))
        
        return self._all_products_cache[z]
            
    def enumerate_alg_integers(self, n):
        """
        Given n, it will list all algebraic integers
           -n   -  n    * u, 
           -n   - (n-1) * u,
           ...
           -n   + (n-1) * u,
          ...
          (n-1) -  n    * u,
          ...
          (n-1) + (n-1) * u
        where u is sqrt(-D) or (1+sqrt(-D))/2 (if D = 3, 7, 11, 15, 19, 23, ...)
        """
        return [ self.CNF(i + self.algInt * j)
                 for i in range(-n,n)
                 for j in range(-n,n) ]
                 
    def to_quadratic_field_quaternion(self, z):
        """
        Given z in self.CNF, returns the coresponding element in self.quadraticFieldQA
        under the embedding of Q(sqrt(-D)) into the quaternions with coefficients in
        Q(sqrt(D)).
        """

        if not z.parent() == self.CNF:
            raise Exception("%r needs to be in imaginary quadratic number field" % z)

        one, i, j, k = self.quadraticFieldQA.basis()
        re, im = ImagQuadraticFieldUtil.getCoefficientsOfElement(z)
        
        return one * re + i * self.sqrtD * im
        
    def to_algebraic_quaternion(self, z):
        """
        Given z in self.CNF, returns the corresponding element in self.algebraicQA
        under the embedding of Q(sqrt(-D)) into the quaternions with real algebraic
        coefficients.
        """
        if not z.parent() == self.CNF:
            raise Exception("Needs to be in imaginary quadratic number field %r" % z)

        one, i, j, k = self.algebraicQA.basis()
        re, im = ImagQuadraticFieldUtil.getCoefficientsOfElement(z)
        
        return one * re + i * self.sqrtDinAA * im

    @staticmethod
    def getCoefficientsOfElement(z):
        """
        Given an element a + b * sqrt(-D) in Q(sqrt(-D)), returns the
        pair (a,b) of Sage Integer/Rational's.
        """

        coefficients = z.lift().coefficients(sparse = False)
        padding = (2 - len(coefficients)) * [ QQ(0) ]
        return tuple(coefficients + padding)
    
