# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from sage.all import sage_eval, Integer

from .imagQuadraticFieldUtil import *
from .pslMatrix import *

__all__ = ['OrbifoldData']

class OrbifoldData:
    """
    Extract information about an orbifold from a regina container c that was
    created by DirichletDomainTriangulator.

    >>> data = OrbifoldData(c)

    This object has the following properties:
    * triangulation is a triangulation of the orbifold. When unglued face 3 of
      each tetrahedron, it becomes the barycentric subdivision of a fundamental
      polyhedron such that vertex 0 of the tetrahedron corresponds to a vertex
      of the polyhedron, 1 to an edge center, 2 to a face center, 3 to the
      center of the polyhedron.
      In particular, each vertex i goes to vertex i of a different simplex under
      any gluing.
    * num_tets is the number of tetrahedra in that triangulation
    * disc which is a sage Integer indicating we are working in Q(sqrt(-disc))
    * imagQuadraticFieldUtil a corresponding instance of ImagQuadraticFieldUtil
    * matrices is a list of PslMatrix'es, one for each barycentric simplex.
      Such a simplex s is glued to another simplex along face 3 by a
      face-pairing of the corresponding face of the polyhedron and its mate
      face. The list stores the matrix taking the mate face to face adjacent to
      s (or, in other words, the face adjacent to s is glued to the mate face of
      a translate of the fundamental polyhedron by the matrix).
    * orders is a list of of triples of integers, one triple for each
      barycentric simplex with the three integers corresponding to the three
      edges of face 3. Each integer gives the order of the edge as singular
      locus (which is a subset of those edges). The first number in a triple
      corresponds to the edge from vertex 1 to 2, then vertex 2 to 0, then
      0 to 1.
    """

    def __init__(self, container):
        self.container = container
        self.triangulation = container.findPacketLabel('Triangulation')
        self.disc = Integer(container.findPacketLabel('D').text())
        self.imagQuadraticFieldUtil = ImagQuadraticFieldUtil(self.disc)
        self.matrices = sage_eval(
            container.findPacketLabel('Matrices').text(),
            {
                self.imagQuadraticFieldUtil.sqrtMinusDName : self.imagQuadraticFieldUtil.sqrtMinusD,
                'ImagQuadraticFieldUtil' : ImagQuadraticFieldUtil,
                'PslMatrix' : PslMatrix
             })

        self.num_tets = self.triangulation.countTetrahedra()

        self.orders = eval(container.findPacketLabel('Orders').text())
