# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from sage.all import AA, QQ, sqrt, lcm, gcd

from .polyhedronUtils import make_plane_equation_integral

__all__ = ['PoincareBallPoint',
           'PoincareBallQuadraticFieldPoint',
           'PoincareBallAlgebraicPoint']

class PoincareBallPoint(object):
    """
    Base class for a point in the Poincare ball which a PslMatrix can act on.
    It is essentially just a wrapper for quaternion, also holding on to an
    instance of ImagQuadraticFieldUtil.
    To address that the quaternions will have coefficients of different types
    at different stages of the computation, we have subclasses from this base
    class.
    These subclasses implement GetQuaternionAlgebra and
    GetEmbeddingIntoQuaternionAlgebra to indicate in what quaternion algebra
    the quaternion making up the point in the Poincare ball is supposed to be
    in, respectively, how convert an element in Q(sqrt(-D)) to that quaternion
    algebra.
    """
    def __init__(self, imagQuadraticFieldUtil, quaternion):
        self._imagQuadraticFieldUtil = imagQuadraticFieldUtil
        self._quaternion = quaternion

        if not quaternion.parent() == self.GetQuaternionAlgebra():
            raise Exception("Given quaternion of wrong type")

    def __hash__(self):
        return hash(self._quaternion)
    
    def __eq__(self, other):
        return self._quaternion == other._quaternion

    def __str__(self):
        return '%s in %s' % (self._quaternion, self.GetQuaternionAlgebra())

class PoincareBallQuadraticFieldPoint(PoincareBallPoint):
    """
    A point in the Poincare ball whose quaternion has coefficients in the
    number field Q(sqrt(-D)). More precisely, the quaternion is an element in
    ImagQuadraticFieldUtil's quadraticFieldQA.
    """

    def GetQuaternionAlgebra(self):
        return self._imagQuadraticFieldUtil.quadraticFieldQA

    def GetEmbeddingIntoQuaternionAlgebra(self):
        return self._imagQuadraticFieldUtil.to_quadratic_field_quaternion

    @staticmethod
    def origin(imagQuadraticFieldUtil):
        """
        Gives the origin of the Poincare Ball model. Needs an instance of
        ImagQuadraticFieldUtil.
        """

        return PoincareBallQuadraticFieldPoint(
            imagQuadraticFieldUtil,
            imagQuadraticFieldUtil.quadraticFieldQA(0))

    def toScaledRationalBisectingPlaneEquationInKleinModel(self):
        """
        This method assumes that the quaternion if of the form
        r + t * j + u * sqrt(D) * k where r, t, and u are rational.
        By a lemma of the paper this is the case for images of the origin
        under PSL(2, Q(sqrt(-D))).

        The method returns an equation for the plane that is bisecting the
        origin and this point when converted to the Klein model - scaled
        such that all coefficients are integers.
        """

        r, s, t, u = self._quaternion.coefficient_tuple()
        r = QQ(r)
        # Ignore s
        t = QQ(t)
        u = QQ(u / self._imagQuadraticFieldUtil.sqrtD) # Scaled to be rational
        
        length = r**2 + t ** 2 + self._imagQuadraticFieldUtil.D * u ** 2

        if length == 0:
            raise Exception(
                "Trying to find bisecting plane for origin %r" % self)

        # Turn equation with rational coefficients into equation with
        # integral coefficients so that it is suitable for
        # Sage's polyhedron.
        return make_plane_equation_integral([length, -r, -t, -u])

    def __repr__(self):
        return "PoincareBallQuadraticFieldPoint(%r, %r)" % (
            self._imagQuadraticFieldUtil, self._quaternion)

class PoincareBallAlgebraicPoint(PoincareBallPoint):
    """
    A point in the Poincare ball whose quaternion has arbitrary real algebraic
    coefficients. More precisely, the quaternion is an element in Sage's
    QuaternionAlgebra over AA, Sage's field of real algebraic numbers.
    """

    def GetQuaternionAlgebra(self):
        return self._imagQuadraticFieldUtil.algebraicQA

    def GetEmbeddingIntoQuaternionAlgebra(self):
        return self._imagQuadraticFieldUtil.to_algebraic_quaternion

    @staticmethod
    def fromScaledRationalPointInKleinModel(imagQuadraticFieldUtil, scaledRationalPt):
        """
        Constructs a point in the Poincare model from a point in the Klein model
        whose coordinates are assumes to be (r, t, u / sqrt(D)) where r, t, u are
        rational.

        This is the counter-part to
        toScaledRationalBisectingPlaneEquationInKleinModel
        in the sense that we need to call this method on the vertices of a
        polyhedron constructed using plane equations from
        toScaledRationalBisectingPlaneEquationInKleinModel.        
        """

        one, i, j, k = imagQuadraticFieldUtil.algebraicQA.basis()

        # Scale the given data so that we get coordinates in the Klein model
        # while converting everything to the base type of the quaternion
        # algebra.
        pt = [ AA(coord) for coord in scaledRationalPt ]
        pt[2] = pt[2] / imagQuadraticFieldUtil.sqrtDinAA
        
        # Compute the square of the (Euclidean distance) to the origin
        dist_sqr = sum([coord**2 for coord in pt])
        
        if dist_sqr > 1:
            raise Exception("Point outside Klein model: %r %r %r" % (
                scaledRationalPt, pt, dist_sqr))
        
        # Factor to go from Klein model to Poincare model
        f = AA(1 / (1 + sqrt(1 - dist_sqr)))
                
        return PoincareBallAlgebraicPoint(
            imagQuadraticFieldUtil,
            f * (pt[0] * one + pt[1] * j + pt[2] * k))
