# Copyright 2017 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from .imagQuadraticFieldUtil import *
from .poincareBall import *

from sage.all import QQ

__all__ = ['PslMatrix']

class PslMatrix:
    """
    Represents an element m in PSL(2, Q(sqrt(-D))) or PSL(2, O) where
    O is the ring of integers of Q(sqrt(-D)).

    It has a method modZSignature() which can be used to detect whether two
    matrices in PSL(2, O) are the same in SL(2, O/<z>)/{+1/-1}.
    More precisely, the method returns a hash-able python object representing
    the image of m in SL(2, O/<z>)/{+1/-1} when m in PSL(2, O).

    An object can be constructed with
    PslMatrix(ImagQuadraticFieldUtil(D), [[a, b], [c, d]]) where
    a, b, c, d are in Q(sqrt(-D)).
    More precisely, D is supposed to be a Sage Integer and
    a, b, c, and d Sage Integer/Rational's or elements in
    ImagQuadraticFieldUtil(D).CNF.

    PslMatrix(ImagQuadraticFieldUtil(D), [[-a, -b], [-c, -d]]) will result in
    an object that equally compares (==) to and gives the same hash as the above
    one.

    Through the act method, this matrix can also act on the Poincare
    ball model. To avoid having non-identity matrices fixing the origin of these
    models, the action is actually conjugated by perturbation_matrix().

    There is some consideration in which number fields the coordinates of points
    in the Poincare model can be, see the paper and poincareBall.py
    for details.
    """

    # _raw indicates that it is performing matrix operations where the matrix
    # is represented as [[a,b],[c,d]]
    @staticmethod
    def _raw_inverse(data):
        return (( data[1][1],-data[0][1]),
                (-data[1][0], data[0][0]))

    @staticmethod
    def _raw_mult(data1, data2):
        def entry(i, j):
            return data1[i][0] * data2[0][j] + data1[i][1]*data2[1][j]
        return ((entry(0,0), entry(0,1)),
                (entry(1,0), entry(1,1)))                 

    @staticmethod
    def _raw_mult3(data1, data2, data3):
        """
        Multiply three matrices.
        """
        return PslMatrix._raw_mult(data1, PslMatrix._raw_mult(data2, data3))

    @staticmethod
    def _raw_conjugate(data1, data2):
        """
        Compute data2 * data1 * data2^-1.
        """
        return PslMatrix._raw_mult3(
            data2, data1, PslMatrix._raw_inverse(data2))

    @staticmethod
    def _raw_apply(data, f):
        """
        Applies f to each entry in the matrix.
        """

        return ((f(data[0][0]), f(data[0][1])),
                (f(data[1][0]), f(data[1][1])))

    @staticmethod
    def _raw_act(data, pt):
        """
        A matrix [[a,b],[c,d]] acts on a point z by
        (a * z + b) * (c * z + d)^-1.

        Note that since z can be a quaternion, the order of the
        multiplication matters.
        """

        num   = data[0][0] * pt + data[0][1]
        denom = data[1][0] * pt + data[1][1]
        
        return num * (denom ** -1)
    
    @staticmethod
    def _raw_upper_half_space_to_poincare_ball_matrix(
        quaternionAlgebra):
        """
        Gives the matrix whose action on quaternions takes the
        upper half space to the unit ball, coefficients will be in the
        given quaternion algebra.
        """
        
        one, i, j, k = quaternionAlgebra.basis()
        return ((one, -j), (one, j))

    @staticmethod
    def _raw_poincare_ball_to_upper_half_space_matrix(
        quaternionAlgebra):
        """
        Gives the matrix whose action on quaternions takes the
        unit ball to the upper half space, coefficients will be in the
        given quaternion algebra.
        """
        
        one, i, j, k = quaternionAlgebra.basis()
        return ((one, one), (j, -j))        
    
    def _raw_perturbation_matrix(self):
        """
        We conjugate the action on the Poincare model by this matrix - which
        depends on the discriminant D.
        The result is [[a,b],[c,d]] with a, b, c, d in Q(sqrt(-D)), i.e.,
        in ImagQuadraticField(D).CNF.
        """

        if self._imagQuadraticFieldUtil.D in (1, 3):
            # For the Gaussian and Eisenstein integers,
            # the Bianchi group PSL(2, O) has rotations about the vertical axis
            # containing the origin. We need to translate a bit horizontally.
            return PslMatrix._raw_apply(
                ((1, self._imagQuadraticFieldUtil.sqrtMinusD / 5), (0, 1)),
                self._imagQuadraticFieldUtil.CNF)

        # No other Bianchi group has such rotations, but it does have rotations
        # about non-vertical axis containing the origin (in the Poincare ball
        # model, it is the point above 0 of height 1 in upper half space model).
        # We need to move it down a tiny amount.
        amount = QQ(10) / QQ(11)

        # The matrix does not have determinant 1, but since we only care about
        # the result of the action on a point, that's ok.
        return PslMatrix._raw_apply(
            ((amount, 0), (0, 1)),
            self._imagQuadraticFieldUtil.CNF)

    def _raw_act_poincare_ball_quaternion(self,
                                          quaternion,
                                          quaternionAlgebra,
                                          embeddingIntoQuaternionAlgebra):
        """
        Let the matrix
            u_to_ball * pert * self * pert^-1 * ball_to_u
        act on the given quaternion where
         * u_to_ball and ball_to_u are the matrices converting
         * pert is the perturbation matrix
         * and self is this PSL(2,Q(sqrt(-D))) matrix itself.

        quaternion is assumed to be in the given quaternionAlgebra with
        embeddingIntoQuaternionAlgebra a function converting an element in
        Q(sqrt(-D)) to that quaternionAlgebra.
        quaternionAlgebra and embeddingIntoQuaternion are used to make the
        above matrix have coefficients in the quaternionAlgebra so that it can
        act on the quaterion.
        """

        # Compute pert * self * pert^-1 with coefficients in Q(sqrt(-D))
        # (such a matrix can act on Q(sqrt(-D)) which embeds into the boundary
        # of the upper half space model.)
        perturbed_action_on_boundary_of_upper_half_space = (
            PslMatrix._raw_conjugate(
                self._data,
                self._raw_perturbation_matrix()))

        # Convert that matrix to have coefficients in the desired quaternion
        # algebra.
        perturbed_action_on_upper_half_space = PslMatrix._raw_apply(
            perturbed_action_on_boundary_of_upper_half_space,
            embeddingIntoQuaternionAlgebra)
        
        # Compute u_to_ball with coefficients in the quaternion algebra
        upper_half_space_to_poincare_ball_matrix = (
            PslMatrix._raw_upper_half_space_to_poincare_ball_matrix(
                quaternionAlgebra))

        # Compute ball_to_u
        poincare_ball_to_upper_half_space_matrix = (
            PslMatrix._raw_poincare_ball_to_upper_half_space_matrix(
                quaternionAlgebra))

        # Compute produce of all these matrices
        action_on_poincare_ball = PslMatrix._raw_mult3(
            upper_half_space_to_poincare_ball_matrix,
            perturbed_action_on_upper_half_space,
            poincare_ball_to_upper_half_space_matrix)

        # And let act on quaternion
        return PslMatrix._raw_act(action_on_poincare_ball, quaternion)

    def act(self, pt):
        """
        Let this matrix act on a PoincareBallQuadraticFieldPoint or
        PoincareBallAlgebraicPoint.
        """

        if not isinstance(pt, PoincareBallPoint):
            raise Exception("PslMatrix cannot act on %r" % pt)
            
        # This will be PoincareBallQuadraticFieldPoint or
        # PoincareBallAlgebraicPoint depending on the input.
        PoincareBallPointType = type(pt)

        # Act on quaternion and construct a PoincareBallQuadraticFieldPoint or
        # PoincareBallAlgebraicPoint from it.
        return PoincareBallPointType(
            self._imagQuadraticFieldUtil,
            self._raw_act_poincare_ball_quaternion(
                pt._quaternion,
                pt.GetQuaternionAlgebra(),
                pt.GetEmbeddingIntoQuaternionAlgebra()))

    @staticmethod
    def _raw_first_sign(data):
        """
        There are two matrices for an element in PSL differing by changing
        signs.
        Multiply a matrix by the sign returned by this function to get the
        canonical representative of the element in PSL.

        More precisely, given
            [[a0 + a1 * sqrt(-D), a2 + a3 * sqrt(-D)], [..., ...]],
        this function returns the sign (i.e., +1 or -1) of the first non-zero
        a0, a1, a2, or a3.
        """
        for entry in data[0]:
            for c in ImagQuadraticFieldUtil.getCoefficientsOfElement(entry):
                if c > 0:
                    return +1
                if c < 0:
                    return -1

        raise Exception("All entries of first row are zero %r.", data)

    def __init__(self, imagQuadraticFieldUtil, data):
        self._imagQuadraticFieldUtil = imagQuadraticFieldUtil

        # Make sure every entry is in the same Sage number field.
        cnf_data = PslMatrix._raw_apply(data, imagQuadraticFieldUtil.CNF)

        # Multiply given matrix by sign to make it the canonical representative
        # of the element in PSL
        first_sign = PslMatrix._raw_first_sign(cnf_data)
        self._data = (
            (first_sign * cnf_data[0][0], first_sign * cnf_data[0][1]),
            (first_sign * cnf_data[1][0], first_sign * cnf_data[1][1]))

    @staticmethod
    def identity(imagQuadraticFieldUtil):
        """
        Gives the identity matrix in PSL(2, Q(sqrt(-D))).
        D is determined from the given instance of ImagQuadraticFieldUtil.
        """

        return PslMatrix(imagQuadraticFieldUtil,
                         ((1,0),(0,1)))

    def __mul__(self, other):
        return PslMatrix(
            self._imagQuadraticFieldUtil,
            PslMatrix._raw_mult(self._data, other._data))

    def __eq__(self, other):
        return self._data == other._data

    def __hash__(self):
        return hash(self._data)

    def inverse(self):
        """
        Inverts the matrix. Assuming the user has given a matrix with
        determinant 1.
        """

        return PslMatrix(
            self._imagQuadraticFieldUtil,
            PslMatrix._raw_inverse(self._data))
            
    def __repr__(self):
        return "PslMatrix(%r, %r)" % (
            self._imagQuadraticFieldUtil, self._data)

    def modIdealSignature(self, ideal):
        """
        Let O be the ring of integers of Q(sqrt(-D)) and I be an ideal in O
        (given as instance ideal of the class Ideal).
        
        This method returns a hash-able object such that two results compare equal
        if and only if the image of this matrix is the same under
            PSL(2, O) -> SL(2, O/I) / {-1/+1}
        (this method assumes this matrix is in PSL(2,O)).

        The result is actually a nested tuple of elements in O.
        """

        signs = [ self._imagQuadraticFieldUtil.CNF(s) for s in [-1, +1]]

        signatures = [
            ((ideal.modulo(s * self._data[0][0]),ideal.modulo(s * self._data[0][1])),
             (ideal.modulo(s * self._data[1][0]),ideal.modulo(s * self._data[1][1])))
            for s in signs ]

        return min(*signatures)

    def isUpperTriangular(self):
        """
        True if upper triangular matrix.
        """
        return self._data[1][0] == 0
