# Copyright 2018 by Matthias Goerner, enischte@gmail.com.
#
# This code is released under the `GNU General Public License, version 2
# <http://www.gnu.org/licenses/gpl-2.0.txt>`_ or (at your option) any
# later version as published by the Free Software Foundation.

from bianchiOrbifold import ImagQuadraticFieldUtil, PslMatrix

from sage.all import sqrt

import regina

__all__ = ['PrincipalCongruenceManifoldGenerator']

def _formatGenerator(D, z):
    a, b = ImagQuadraticFieldUtil.getCoefficientsOfElement(z)
    return '%r_plus_%r_sqrt_minus_%d' % (
        a if a.is_integer() else float(a),
        b if b.is_integer() else float(b),
        D)
    
def _formatIdeal(ideal):
    return '_and_'.join(
        [ _formatGenerator(ideal.D, gen) for gen in ideal.gens ])

_edgeNumber = [ regina.Edge3.edgeNumber[(j+1)%3][(j+2)%3] for j in range(3) ]

class PrincipalCongruenceManifoldGenerator:
    """
    Creates the manifold obtained as quotient by the principal congruence group
              ker(PSL(2, O) -> SL(2, O/I) / {+1/-1})
    where O is the ring of integers of Q(sqrt(-D)) and I an ideal of type
    bianchiOrbifold.Ideal.

    The constructor takes a bianchiOrbifold.OrbifoldDescription for the
    respective Bianchi orbifold and two Sage Integer/Rational's a and b.

    The resulting object has the following properties:
     * triangulation is a Regina triangulation of the principal congruence
       manifold. It is a cover of the triangulation given for the Bianchi
       orbifold. If the principal congruence group has torsion, it will be
       a triangulation of the topological manifold underlying the quotient
       orbifold.
     * baseName is a suggested name for files related to this (principal)
       congruence group.

    The triangulation is a cover of the Bianchi orbifold, namely it is obtained
    by gluing together copies of the fundamental polyhedron, one for each
    element in Im(PSL(2, O) -> SL(2, O/I) / {+1/-1}).

    If makeNanotube = True is supplied, you get a only the cusp neighborhood
    of one cusp of the principal congruence manifold. More precisely, it is
    considering only those copies of the fundamental polyhedron that correspond
    to upper unit triangular matrices in
    Im(PSL(2, O) -> SL(2, O/I) / {+1/-1}).

    If is_gamma1 = True is supplied, the congruence manifold
                 (PSL(2, O) -> SL(2, O/I) / {+1/-1})^-1 (P)
    where P are the unit upper triangular matrices in SL(2, O/I) / {+1/-1} is
    generated instead.

    """

    # Assume the fundamental polyhedron of the Bianchi orbifold is build from
    # k (barycentric) simplicies.
    #
    # The i-th copy of the fundamental polyhedron in the triangulation of the
    # consists of simplicies i * k + j with j = 0, ..., k - 1 such that simplex
    # i * k + j corresponds to simplex j in the Bianchi orbifold.
    #
    # We store the PslMatrix taking the base copy 0 to the i-th copy in
    # self.matrices[i]. self.matrix_to_fund_index is the inverse mapping.
    #
    # The triangulation, self.matrices, and self.matrix_to_fund_index will be
    # build up progressively.

    def _find_or_add(self, pslMatrix):
        """
        Compute the image of pslMatrix under PSL(2, O) -> SL(2, O/<z>)/{+1/-1}.
        If a fundamental polyhedron for this image exists, return it, otherwise,
        create a new one and update self.matrices and self.matrix_to_fund_index.
        """

        # Image under PSL(2, O) -> SL(2, O/I)/{+1/-1}
        if self.is_gamma1:
            mSig = pslMatrix.modIdealSignature(self.ideal)[0]
        else:
            mSig = pslMatrix.modIdealSignature(self.ideal)

        # Determine whether fundamental polyhedron for it already exists
        i = self.matrix_to_fund_index.get(mSig)
        if not i is None:
            # It exists and it is the i-th copy
            return i

        # Create a new copy with index i
        i = self.triangulation.countTetrahedra() / self.k

        # Book-keeping
        self.matrices.append(pslMatrix)
        self.matrix_to_fund_index[mSig] = i

        # Create k new simplicies
        new_simplices = [ self.triangulation.newSimplex() for j in range(self.k) ]

        # Perform the internal gluing for the new copy of the fundamental
        # polyhedron using the orbifold as guide
        #
        # For each new simplex and the corresponding simplex in the orbifold
        for orbSimp, newSimp in zip(
               self.orbifoldData.triangulation.simplices(), new_simplices):
            # For each internal face
            for f in range(3):
                # Find the neighboring simplex in the orbifold
                neighborOrbSimp = orbSimp.adjacentSimplex(f)
                # Find the corresponding new simplex
                neighborNewSimp = new_simplices[neighborOrbSimp.index()]
                # Glue the two new simplices
                newSimp.join(f, neighborNewSimp, regina.NPerm4())

        # Return index of new polyhedron
        return i

    def __init__(self, orbifoldData, ideal, makeNanotube = False, is_gamma1 = False):

        self.orbifoldData = orbifoldData
        self.ideal = ideal

        self.is_gamma1 = is_gamma1

        self.k = self.orbifoldData.num_tets

        self.triangulation = regina.Triangulation3()
        
        self.matrices = [ ]
        self.matrix_to_fund_index = { }

        # Create the base copy for the identity matrix
        identity = PslMatrix.identity(self.orbifoldData.imagQuadraticFieldUtil)
        self._find_or_add(identity)

        # The index of the simplex we will process next
        n = 0
        
        # While there are unprocessed simplices
        while n < self.triangulation.countTetrahedra():
            # j as above
            j = n % self.k
            pairing_matrix = self.orbifoldData.matrices[j]

            if not makeNanotube or pairing_matrix.isUpperTriangular():
                # i as above
                i = n / self.k
                
                # matrix of this fundamental polyhedron
                m = self.matrices[i]

                # matrix for "other" fundamental polyhedron - the one
                # glued along the face corresponding to simplex n
                mOther = m * pairing_matrix
                # index of other fundamental polyhedron
                iOther = self._find_or_add(mOther)

                # simplex in orbifold corresponding to this simplex
                orbSimplex = self.orbifoldData.triangulation.simplex(j)
                jOther = orbSimplex.adjacentTetrahedron(3).index()
                nOther = iOther * self.k + jOther

                simp = self.triangulation.simplex(n)
                otherSimp = self.triangulation.simplex(nOther)

                # Glue this simplex to the appropriate simplex in
                # other fundamental polyhedron
                simp.join(3, otherSimp, regina.NPerm4())

            # Advance to next simplex
            n += 1

        if self.is_gamma1:
            self.baseName = 'gamma1_' + _formatIdeal(ideal)
        else:
            self.baseName = 'pSL_' + _formatIdeal(ideal)
        self.isOrbifold = self._is_orbifold()
        if self.isOrbifold:
            self.baseName += '_orbifold'

        self.triangulation.setLabel(self.baseName)

    def _is_orbifold(self):
        if self.is_gamma1:
            n = self.triangulation.size()
        else:
            n = self.orbifoldData.triangulation.size()

        for i in range(n):
            iOrb = i % self.orbifoldData.triangulation.size()

            for j in range(3):
                e = _edgeNumber[j]
                order = self.orbifoldData.orders[iOrb][j]

                orb_edge_degree = (
                    self.orbifoldData.triangulation.simplex(iOrb)
                    .edge(e).degree())
                edge_degree = (
                    self.triangulation.simplex(i)
                    .edge(e).degree())

                if edge_degree < order * orb_edge_degree:
                    return True

                if edge_degree > order * orb_edge_degree:
                    raise Exception("Something is wrong")

        return False

            
