import regina
from regina import Triangulation3, Perm4
from collections import deque

__all__ = ['EdgeCollapsingEngine']

class EdgeCollapsingEngine:
    """
    Given a regina triangulation, this class will collapse edges
    (preferring the ones with higher degrees) whenever collapsing the
    edge does not change the topology.

    The method of checking whether collapsing an edge does not change
    the topology is similar to Regina's collapseEdge. However, Regina's
    implementation is invalidating and recomputing the entire skeleton
    of the triangulation when collapsing a single edge, so it does
    not scale to larger triangulations.

    Here, we instead only invalidate the edge classes near the collapsed
    edge and merge the two vertices where the collapsed edge ended.

    There is a high-level static method
    EdgeCollapsingEngine.reginaTrigCollapsedEdges to collapse all
    edges of a regina triangulation.
    """

    @staticmethod
    def reginaTrigCollapsedEdges(reginaTrig):
        """
        Takes a regina triangulation and returns a new regina triangulation
        where all possible edges have been collapsed.
        """

        e = EdgeCollapsingEngine(reginaTrig)
        e.collapseEdges()
        return e.getReginaTrig()

    def __init__(self, reginaTrig):
        
        # Don't actually store the regina triangulation.
        # Instead, just store the adjacent tetrahedra and gluings.

        # For each tetrahedron and face, store index of adjacent tet.
        self.adjacentTets = [
            [ t.adjacentTetrahedron(i).index() for i in range(4) ]
            for t in reginaTrig.tetrahedra() ]
        # For each tetrahedron and face, store gluing permutation
        # (it is a regina.Perm4).
        self.adjacentGluings = [
            [ t.adjacentGluing(i) for i in range(4) ]
            for t in reginaTrig.tetrahedra() ]

        # For each tetrahedron and each of the six edges, store
        # the edge class.
        # This is initialized with None here.
        # But eventually for edges of tetrahedra that become identified in the
        # triangulation, the corresponding array entries will be references
        # to the same python object.
        # The edges are indexed according to regina.Edge3.edgeNumber.
        self.edges = [ [ None for i in range(6) ]
                       for t in reginaTrig.tetrahedra() ]
        # For each tetrahedron and each of the four vertices, store the
        # vertex class.
        # Similar notes apply.
        self.vertices = [ [ None for i in range(4) ]
                          for t in reginaTrig.tetrahedra() ]
        # An array with the i-th entry being a double-ended queue (deque) of
        # all unprocessed edge classes of order i. It is a deque so that we can
        # remove an edge class we are about to process from the front and
        # append the new edge classes at the back.
        # It really acts as priority queue with the priority being the degree
        # of an edge class. But since the number of distinct priorities is
        # fairly low (something like a dozen), we use an array of deques.
        #
        # Note that python lists are 0-indexed but an edge class has degree
        # at least 1, so we substract 1 from the degree to get the index.
        #
        # During the loop to collapse edges, an edge will be removed
        # from this list and then checked whether it can be collapsed
        # without changing the topology.
        #
        # Note that an edge class will be invalidated when a nearby
        # edge was collapsed and new edge classes for the nearby edges
        # will be created.
        #
        # Edge classes that get invalidated won't be removed from this
        # list (only the while-loop to find collapsable edges is removing
        # them). Instead, edge classes have a flag indicating whether they
        # have been removed.
        #
        # This means that if an edge e was not collapsable, but collapsing a
        # nearby edge makes it collapsable, the old edge class for e gets
        # invalidated and a new edge class for e will be added so that it
        # can be picked up by the while-loop. The degree of the edge e might
        # have increased, so the old edge class and new edge class will be
        # stored in different places in the edgesByDegree array.
        self.edgesByDegree = []

        # Initialize the above structures.
        self._initEdgesAndVertices()

    def collapseEdges(self):
        """
        While there is an unprocessed edge class, pick one with highest degree
        and collapse it if this does not change the topology and the edge class
        has not been invalidated.
        """
        while self.collapseNextEdge():
            pass

    def collapseNextEdge(self):
        """
        Get an unprocessed edge class of highest degree (returning False
        if none exists, otherwise True) and remove from list of unprocessed
        edges. If it has not been invalidated and collapsing it does not
        change the topology, collapse it.
        """

        edge = self.getNextEdge()
        if not edge:
            return False
        self.collapseEdge(edge)
        return True

    def _initEdgesAndVertices(self):
        """
        Init edge and vertex classes.
        """

        # Go through each edge of each tetrahedron
        for tet in range(len(self.adjacentTets)):
            for i in range(6):
                # If it has not been assigned an edge class yet.
                if not self.edges[tet][i]:
                    # Create edge class,
                    # find all edges of tets around this triangulation edge,
                    # assign to the edges of the tetrahedra,
                    # put edge class into edgesByDegree list.
                    self.addEdge(Edge.addToEngine(self, tet, i))

        # Go through each vertex of each tetrahedron
        for tet in range(len(self.adjacentTets)):
            for i in range(4):
                # If it has not been assigned a vertex class yet.
                if not self.vertices[tet][i]:
                    # Create vertex class,
                    # recurse to find all vertices and tets belonging
                    # to this vertex class,
                    # assign to vertices of tetrahedra.
                    Vertex.addToEngine(self, tet, i)

    def addEdge(self, edge):
        """
        Add edge class to edgesByDegree list.
        """
        degree = len(edge.embeddings)
        # If necessary, length edgesByDegre list.
        for i in range(degree - len(self.edgesByDegree)):
            self.edgesByDegree.append(deque())

        # Append to respective degree.
        self.edgesByDegree[degree - 1].append(edge)

    def getNextEdge(self):
        """
        Gets an unprocessed edge class of highest degree. None if no
        unprocessed edge class.
        """

        if not self.edgesByDegree:
            return None
        # Get edge class
        edge = self.edgesByDegree[-1].popleft()
        # Shorten edgesByDegree if necessary so that the highest entry
        # isn't an empty deque and we can call this method again.
        for i in xrange(len(self.edgesByDegree) -1, -1, -1):
            if self.edgesByDegree[i]:
                break
            del self.edgesByDegree[i]
        return edge

    def collapseEdge(self, edge):
        """
        Collapse the edge in the triangulation corresponding to the given edge
        class - if this does not change the topology and the edge class was not
        invalidated.

        This is a reimplementation of regina's collapseEdge. It might be helpful
        to read through some of the comments in the regina source code to follow
        along.

        However, unlike regina, this method carefully only invalidates the edge
        classes of nearby edges and merges the vertices.
        """

        if edge.isDead():
            # Do nothing if edge was already invalidated.
            return

        # Get the embeddings of the edge class. Similar to regina, this is a
        # list of pairs (tetrahedron index, permutation).
        # The edge between vertex perm[0] and perm[1] of the tetrahedron
        # belongs to the edge class.
        # perm[2] and perm[3] are chosen consistently with the orientation.
        embeddings = edge.embeddings

        # Determine whether there is a cycle of bigons. This is documented in
        # detail in the corresponding regina source code.
        # The bigons come from collapsing the triangles touching the edge in
        # the triangulation. In term of the (tetrahedron index, permutation)
        # pair, the two edges are the ones from vertex perm[0] to perm[2] and
        # from vertex perm[1] to perm[2].
        # Each bigon gives a pair of edge classes. We get a list of pairs
        # of edge classes.
        # This forms a graph where a node is an edge class and an "edge" is
        # a bigon. Check whether the graph has a cycle.
        if hasCycle(
            [ 
                [ self.edges[tet][regina.Edge3.edgeNumber[perm[i]][perm[2]]]
                  for i in range(2) ]
                for tet, perm in embeddings ]):
            return

        # Collapsing the tetrahedra gives triangles and we need to determine
        # whether there is a cycle of those. Again, this is documented in
        # detail in the corresponding regina source code.

        # Imagine the edge is vertical. Each tetrahedron about the edge
        # has a top and a bottom face.
        #
        # We again get a graph where the nodes are face classes and we have
        # an "edge" between the top and bottom face of each tetrahedron.
        trianglePairs = []

        for tet, perm in embeddings:
            # Get face classes of bottom and top face of tetrahedron.
            #
            # A face class has exactly two faces which are pairs
            # (tet index, face).
            # Pick as canonical representative of the face class the
            # lexicographically smallest pair.
            #
            tet0 = self.adjacentTets[tet][perm[0]]
            tet1 = self.adjacentTets[tet][perm[1]]
            perm0 = self.adjacentGluings[tet][perm[0]]
            perm1 = self.adjacentGluings[tet][perm[1]]
            
            face0 = perm0[perm[0]]
            face1 = perm1[perm[1]]

            trianglePairs.append(
                [ min((tet, perm[0]), (tet0, face0)),
                  min((tet, perm[1]), (tet1, face1)) ])

        if hasCycle(trianglePairs):
            return

        # Pick one embedding to get the two vertex classes of the edge.
        tet, perm = embeddings[0]

        v0 = self.vertices[tet][perm[0]]
        v1 = self.vertices[tet][perm[1]]

        # The two vertex clases must be distinct
        if AreVerticesIdentified(v0, v1):
            return

        # Get the Euler characteristics
        eulerChar0 = v0.getEulerChar()
        eulerChar1 = v1.getEulerChar()

        # We cannot collapse an edge between two ideal vertices.
        if eulerChar0 == 0 and eulerChar1 == 0:
            return

        # Remember all tests touching the edge we are collapsing
        # as well as their neighbors.
        # We will invalidate and recompute all those edge classes.
        all_tets = set()

        # Go around the edge
        for tet, perm in embeddings:
            # Look at the tetrahedron glued to the bottom and to the top
            # face
            tet0 = self.adjacentTets[tet][perm[0]]
            tet1 = self.adjacentTets[tet][perm[1]]
            perm0 = self.adjacentGluings[tet][perm[0]]
            perm1 = self.adjacentGluings[tet][perm[1]]
            
            face0 = perm0[perm[0]]
            face1 = perm1[perm[1]]
            
            # Determine the permutations that we need to apply to
            # glue the tetrahedron glued to our bottom face directly
            # to the tetrahedron glued to our top face.
            perm1To0 = (perm1 *
                        perm *
                        Perm4(1,0,2,3) *
                        perm.inverse() *
                        perm0.inverse())

            perm0To1 = perm1To0.inverse()

            # Glue the above two tetrhaedron tgoether
            self.adjacentTets[tet0][face0] = tet1
            self.adjacentTets[tet1][face1] = tet0
            
            # Using the appropriate permutations
            self.adjacentGluings[tet0][face0] = perm1To0
            self.adjacentGluings[tet1][face1] = perm0To1

            all_tets.add(tet)
            all_tets.add(tet0)
            all_tets.add(tet1)

        # Mark the tetrahedra about the collapsed edge as dead.
        # Thus, we will skip them when recomputing the edge classes
        # - and later when generating the regina triangulation.
        for tet, perm in embeddings:
            for i in range(4):
                self.adjacentTets[tet][i] = -1

        # Make code deterministic
        all_tets = sorted(all_tets)

        # Invalidate all edge classes of all the tetrahedra we touched
        for tet in all_tets:
            for i in range(6):
                self.edges[tet][i].markDead()
                self.edges[tet][i] = None
        
        # Recompute the edge classes for the non-dead tetrahedra.
        for tet in all_tets:
            if self.adjacentTets[tet][0] != -1:
                for i in range(6):
                    if not self.edges[tet][i]:
                        self.addEdge(Edge.addToEngine(self, tet, i))

        # Merge the vertices.
        IdentifyVertices(v0, v1)

    def getReginaTrig(self):
        """
        Generate regina triangulation from internal representation.
        """

        T = Triangulation3()

        # Make a regina simplex for each non-dead simplex.
        # Fill in None for dead-simplices.
        tets = [ T.newSimplex() if adjacentTets[0] != -1 else None
                 for adjacentTets in self.adjacentTets ]

        # For each simplex, dead or alive
        for tet, adjacentTets, adjacentGluings in zip(
            tets, self.adjacentTets, self.adjacentGluings):

            # For the four faces of each simplex
            for f, (adjacentTet, adjacentGluing) in enumerate(zip(
                adjacentTets, adjacentGluings)):

                # Skip dead simplices
                if adjacentTet != -1:
                    if not tets[adjacentTet]:
                        print "BAD"

                    # Transfer adjacency information to regina.
                    tet.join(f, tets[adjacentTet], adjacentGluing)

        return T

# A list of 6 even permutations for the 6 edges of a tetrahedron to start
# finding the embeddings of an edge class.
#
# The i-th permutation is such that the i-th edge of a tetrahedron
# goes from perm[0] to perm[1].
# 
_EdgePerm = [
    [ p for p in Perm4.orderedS4
      if p[0] == regina.Edge3.edgeVertex[i][0] and
      p[1] == regina.Edge3.edgeVertex[i][1] and
      p.sign() > 0 ][0]
    for i in range(6) ]

# A list of 4 even permutations for the 4 vertices of a tetrahedron to start
# finding the embeddings of a vertex class.
#
# Recall that an embedding of a vertex class is a pair (tet index, perm)
# such that vertex perm[0] of the tet belongs to the vertex class.
#
_VertexPerm = [
    [ p for p in Perm4.orderedS4
      if p[0] == i and p.sign() > 0][0]
    for i in range(4) ]

# Like regina, we use union-find to test whether a graph has a cycle,
# again see unionFindInsert in the regina source code.
#

def getRoot(d, a):
    if a in d:
        return getRoot(d, d[a])
    return a

def hasCycle(listOfPairs):
    """
    listOfPairs is interpreted as a graph, i.e., each element is
    assumed to be a pair of two vertices forming an edge.
    The vertex objects can be any kind of objects that are ordered.

    Returns true if this graph has cycles.
    """
    
    d = {}

    for a, b in listOfPairs:
        if a < b:
            e, f = a, b
        else:
            e, f = b, a

        r1 = getRoot(d, e)
        r2 = getRoot(d, f)

        if r1 == r2:
            return True

        d[r1] = r2

    return False

class Edge:
    """
    An Edge class for the triangulation encoded in an EdgeCollapseEngine.
    """
    @staticmethod
    def getEmbeddings(edgeCollapseEngine, tet, perm):
        """
        Consider the edge of the tetrahedron (given by its index tet) between
        vertex perm[0] and and perm[1]. Find all other embeddings of the edge
        in the triangulation.
        """

        this_tet, this_perm = tet, perm

        result = []

        while True:
            result.append((this_tet, this_perm))
            face = this_perm[2]

            this_tet, this_perm = (
                edgeCollapseEngine.adjacentTets[this_tet][face],
                edgeCollapseEngine.adjacentGluings[this_tet][face] * this_perm * Perm4(0,1,3,2))
            
            if this_tet == tet and this_perm == perm:
                return result
            
    def __init__(self, embeddings):
        # All embeddings of this edge class.
        # Similar to regina, this is a
        # list of pairs (tetrahedron index, permutation).
        # The edge between vertex perm[0] and perm[1] of the tetrahedron
        # belongs to the edge class.
        # perm[2] and perm[3] are chosen consistently with the orientation.
        self.embeddings = embeddings

    @staticmethod
    def addToEngine(edgeCollapseEngine, tet, i):
        """
        For edge i of the tetrahedron with index tet, compute edge class.
        Assign this new edge class to all the edges of all the tetrahedra that
        belong to the new edge class.
        """

        embeddings = Edge.getEmbeddings(edgeCollapseEngine, tet, _EdgePerm[i])

        edge = Edge(embeddings)

        for this_tet, this_perm in embeddings:
            j = regina.Edge3.edgeNumber[this_perm[0]][this_perm[1]]
            edgeCollapseEngine.edges[this_tet][j] = edge

        return edge

    def markDead(self):
        """
        Invalidate edge.
        """
        self.embeddings = []
        
    def isDead(self):
        """
        Was edge invalidated.
        """
        return not self.embeddings

def IdentifyVertices(v0, v1):
    """
    Identify two vertex classes. The arguments need to be
    instances of Vertex.
    """

    v0 = v0.getCanonicalRepresentative()
    v1 = v1.getCanonicalRepresentative()

    eulerChar0 = v0.getEulerChar()
    eulerChar1 = v1.getEulerChar()

    # Identify v0 with v1
    v0._canonicalRepresentative = v1

    # Merging to finite vertices gives a finite vertex.
    if eulerChar0 == 0 or eulerChar1 == 0:
        v1.eulerChar = 0
    else:
        # We only ever merge an ideal vertex with a finite
        # vertex, so new eulerChar is 2.
        v1.eulerChar = 2

def AreVerticesIdentified(v0, v1):
    """
    Returns true if vertex class v0 is identified with vertex class v1.
    The arguments need to be instances of Vertex.
    """

    return v0.getCanonicalRepresentative() is v1.getCanonicalRepresentative()

class Vertex:
    """
    A vertex class. It will store the Euler characteristic and it carries
    data so that we can check with AreVerticesIdentified whether two vertex
    classes were "merged" (with IdentifyVertices).
    """

    def __init__(self, eulerChar):
        self.eulerChar = eulerChar

        # If _canonicalRepresentative is set, it means that this Vertex
        # was identified with another Vertex. If we want to lookup data such
        # as eulerChar, we should ignore what is stored on this vertex and
        # instead consult the data from _canonicalrepresentative. This
        # works recursively.
        self._canonicalRepresentative = None

    def getEulerChar(self):
        return self.getCanonicalRepresentative().eulerChar

    def getCanonicalRepresentative(self):
        if self._canonicalRepresentative:
            # Update canonical representative such that future lookups will be faster
            self._canonicalRepresentative = self._canonicalRepresentative.getCanonicalRepresentative()
            return self._canonicalRepresentative

        return self

    @staticmethod
    def addToEngine(edgeCollapseEngine, tet, i):
        """
        For vertex i of the tetrahedron with index tet, compute vertex class.
        Assign this new vertex class to all the vertex of all the tetrahedra that
        belong to the new vertex class.
        """

        # We really enumerate all oriented flags in the triangulation where the
        # vertex of the flag belongs to the vertex class.
        embeddings = set()

        # Start with one such flag
        pending_embeddings = [ (tet, _VertexPerm[i]) ]

        # Keep track of the number of edges to compute the euler characterisic
        num_edges = 0

        # Breath-first search to find all flags
        while pending_embeddings:
            this_tet, this_perm = pending_embeddings.pop()
            # If we haven't processed this flag before
            if not (this_tet, this_perm.permCode()) in embeddings:
                
                num_edges += 1

                # Get all flags about this edge
                for ntet, nperm in Edge.getEmbeddings(edgeCollapseEngine, this_tet, this_perm):
                    
                    embeddings.add((ntet, nperm.permCode()))

                    # Add the other two flags of the meeting this tetrahedron
                    # at the vertex for future processing.
                    pending_embeddings.append(
                        (ntet, nperm * Perm4(0, 2, 3, 1)))
                    pending_embeddings.append(
                        (ntet, nperm * Perm4(0, 3, 1, 2)))

        # Create vertex class. Euler characteristic can be computed from
        # number of adjacent edges and number of flags
        vertex = Vertex(
            eulerChar = num_edges - len(embeddings) / 6)

        for this_tet, this_perm in embeddings:
            j = Perm4.fromPermCode(this_perm)[0]
            edgeCollapseEngine.vertices[this_tet][j] = vertex

